%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TEMPORAL DEBLURRING SIMULATION
%
% This simulation program show a simple temporal deblurring example of a
% video of the rotating pepper image. This simulation program consists of
% four steps:
%
%   1. LOAD A TEST IMAGE
%      We use a pepper image as a test image in this simulation.
%
%   2. CREATE A SIMULATED GROUND-TRUTH SEQUENCE
%      We create a simulated ground-truth sequence by rotating the pepper
%      image counterclockwise 0.5 degree per frame.
%
%   3. BLUR THE SIMULATED SEQUENCE WITH TEMPORAL PSF AND ADD SOME NOISE
%      The temporal PSF in this example is a 1x1x8 uniform PSF. Using the
%      PSF, we blur the simulated sequence and add weak noise (white
%      Gaussian noise with standard deviation 2). Assuming that the blurred
%      noisy sequence is the output of the space-time interpolator, the
%      next step is deblurring.
%
%   4. DEBLUR THE BLURRED NOISY SEQUENCE
%      The final step is deblurring. We use the true PSF in this example.
%      Using the steepest descent method, we iteratively estimate the
%      underlying sequence with 3-D (digital) total valiation.
%
% Note : An appropriate mirroring around the edges of the 3-D array of the
% videos is also necessary to avoid ringing. To simplify the simulation
% code, the mirroring is omitted from this code.
%
% wrote by Hiroyuki Takeda, Feb 9th, 2011


%%%%%%%%%%%%%%%%%%%%%%%%
% 1. LOAD A TEST IMAGE %
%%%%%%%%%%%%%%%%%%%%%%%%
img = imread('pepper.bmp');
% make the image size odd so that there is a pixel in the center of the
% rotation.
img = double(img(1:end-1, 1:end-1)); % make the image size odd
[N M] = size(img);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2. CREATE A GROUND-TRUTH SEQUENCE %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%th = [zeros(1,8), 0 : 0.5 : 44.5, ones(1,8)*44.5];
%th = [zeros(1,8), 0 : 1 : 89, ones(1,8)*89];
th = -8 : 1 : 97;
cut = 37;
seq_true = zeros(N-cut*2, M-cut*2, length(th));
for i = 1 : length(th)
    % rotate the image instead of rotateing the camera (the 'imrotate'
    % blurs the output image, but we ignore in this simulation)
    imrot = imrotate(img, th(i), 'bicubic', 'crop');
    % cut out the 
    seq_true(:,:,i) = imrot(1+cut:end-cut, 1+cut:end-cut);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 3. BLUR THE SEQUENCE WITH A TEMPORAL PSF %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% temporal PSF (1x1x8 uniform PSF)
G = ones(1, 1, 9);
G(:,:,9) = 0;
G = G / sum(G(:));

% convolve the rotating sequence with the temporal PSF 
seq_blur = convn(seq_true, G, 'same');
seq_true = seq_true(:,:,5:end-4);
seq_blur = seq_blur(:,:,5:end-4);

% add some noise
sigma = 2;
randn('state', 0);
seq_blur = seq_blur + randn(size(seq_blur)) * sigma;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 4. DEBLURRING WITH TOTAL VARIATION %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IT = 400;           % the number of iterations of the steepest descent
step = 0.5;         % the step size of the steepest descent update
lambda = 0.08;      % the regularization parameter
Gt = flipdim(G, 3); % the flipped version of the temporal PSF
u = seq_blur;       % initialize with the blurred sequence

for it = 1 : IT
   
    % the gradient from the likelihood term
    ec = convn(seq_blur - convn(u, G, 'same'), Gt, 'same'); % G^T (y - G x)
    
    % the gradient from the TV regularization term
    % x1-direction
    tmp = sign(u - ishift3(u, 1, 0, 0));
    tv = tmp - ishift3(tmp, -1, 0, 0);
    % x2-direction
    tmp = sign(u - ishift3(u, 0, 1, 0));
    tv = tv + tmp - ishift3(tmp, 0, -1, 0);
    % t-direction
    tmp = sign(u - ishift3(u, 0, 0, 1));
    tv = tv + tmp - ishift3(tmp, 0, 0, -1);
    
    % update
    update = step * (ec - lambda * tv);
    update(:,:,1:9) = 0;
    update(:,:,end-8:end) = 0;
    u = u + update;
    
    % display the estimation status
    figure(1);
    fnum = 40;
    subplot(1,3,1); imagesc(u(:,:,fnum)); colormap(gray); axis image; colorbar;
    title(sprintf('The current estimate of the frame at t=%d, %d iterations', fnum, it));
    subplot(1,3,2); imagesc(ec(:,:,fnum)); colormap(gray); axis image; colorbar;
    title('The gradient of the likelihood term');
    subplot(1,3,3); imagesc(tv(:,:,fnum)); colormap(gray); axis image; colorbar;
    title('The gradient of the total valiation term');
    pause(0.1);
end

% display the result
figure(2);
fnum = 40;
subplot(1,3,1); imagesc(seq_true(:,:,fnum)); colormap(gray); axis image; colorbar;
title(sprintf('The original frame at t = %d', fnum));
subplot(1,3,2); imagesc(seq_blur(:,:,fnum)); colormap(gray); axis image; colorbar;
error = seq_true(:,:,fnum) - seq_blur(:,:,fnum);
psnr = 20 * log10(255 / sqrt(mean(error(:).^2)));
title(sprintf('The simulated blurly noisy frame at t = %d, PSNR = %4.2f [dB]', fnum, psnr));
subplot(1,3,3); imagesc(u(:,:,fnum)); colormap(gray); axis image; colorbar;
error = seq_true(:,:,fnum) - u(:,:,fnum);
psnr = 20 * log10(255 / sqrt(mean(error(:).^2)));
title(sprintf('The deblurred frame at t = %d by our method, PSNR = %4.2f [dB]', fnum, psnr));



